import uvicorn
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

from pandahub.lib.PandaHub import PandaHubError
from pandahub.api.routers import net, projects, timeseries, users, auth

app = FastAPI()

origins = [
    "http://localhost:8080",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(net.router)
app.include_router(projects.router)
app.include_router(timeseries.router)
app.include_router(users.router)
app.include_router(auth.router)


@app.exception_handler(PandaHubError)
async def pandahub_exception_handler(request: Request, exc: PandaHubError):
    return JSONResponse(
        status_code=exc.status_code,
        content=str(exc),
    )


if __name__ == "__main__":
    uvicorn.run("main:app", host="127.0.0.1", port=8002, log_level="info", reload=True)



import requests
import pandapower as pp
import pandas as pd
import numpy as np
from pathlib import Path
import os
import json
from fastapi.encoders import jsonable_encoder

class PandaHubClient:
    def __init__(self):
        config = os.path.join(Path.home(), "pandahub.config")
        try:
            with open(config, "r") as f:
                d = json.load(f)
        except FileNotFoundError:
            raise UserWarning("No pandahub configuration file found - log in first")
        self.url = d["url"]
        self.token = d["token"]
        self.cert = None
        if d.get("client_cert_path") and d.get("client_key_path"):
            self.cert = (d["client_cert_path"], d["client_key_path"])
        self.project_id = None

    def set_active_project(self, project_name):
        r = self._post("/projects/set_active_project", json=locals())
        if r.ok:
            self.project_id = r.json()
        else:
            self.project_id = None
        return r

    ### PROJECT HANDLING

    def create_project(self, name):
        return self._post("/projects/create_project", json=locals())

    def delete_project(self, i_know_this_action_is_final=False,
                             error_on_missing_project=True):
        return self._post("/projects/delete_project", json=locals())

    def project_exists(self, name):
        return self._post("/projects/project_exists", json=locals()).json()

    def get_projects(self):
        return self._post("/projects/get_projects").json()

    def get_project_settings(self):
        return self._post("/projects/get_project_settings").json()

    def set_project_settings(self, settings):
        return self._post("/projects/set_project_settings", json=locals())

    def get_project_metadata(self):
        return self._post("/projects/get_project_metadata").json()

    def set_project_metadata(self, metadata):
        return self._post("/projects/set_project_metadata", json=locals())

    ### NET HANDLING

    def write_network_to_db(self, net, name, overwrite=True):
        json = locals()
        json["net"] = pp.to_json(net)
        return self._post("/net/write_network_to_db", json=json)

    def get_net_from_db(self, name, include_results=True, only_tables=None):
        r = self._post("/net/get_net_from_db", json=locals())
        return pp.from_json_string(r.json())


    ### ELEMENT HANDLING

    def get_net_value_from_db(self, net_name, element, element_index, parameter):
        return self._post("/net/get_net_value_from_db", json=locals()).json()

    def set_net_value_in_db(self, net_name, element, element_index, parameter, value):
        return self._post("/net/set_net_value_in_db", json=locals())

    def create_element_in_db(self, net_name, element, element_index, data):
        return self._post("/net/create_element_in_db", json=locals())

    def create_elements_in_db(self, net_name, element_type, elements_data):
        return self._post("/net/create_elements_in_db", json=locals())

    def delete_net_element(self, net_name, element, element_index):
        return self._post("/net/delete_net_element", json=locals())


    ### TIMESERIES

    def multi_get_timeseries_from_db(self, filter_document={}, timestamp_range=None,
                                    exclude_timestamp_range=None,
                                    global_database=False):
        ts = self._post("/timeseries/multi_get_timeseries_from_db", json=locals()).json()
        for i, data in enumerate(ts):
            ts[i]["timeseries_data"] = pd.Series(json.loads(data["timeseries_data"]))
        return ts

    def get_timeseries_from_db(self, filter_document={}, timestamp_range=None,
                                    exclude_timestamp_range=None,
                                    global_database=False):
        r = self._post("/timeseries/get_timeseries_from_db", json=locals())
        return pd.Series(json.loads(r.json()))

    def write_timeseries_to_db(self, timeseries, data_type, element_type=None,
                               netname=None, element_index=None, name=None,
                               global_database=False, collection_name="timeseries"):
        json = locals()
        json["timeseries"] = json["timeseries"].to_json(date_format="iso")
        return self._post("/timeseries/write_timeseries_to_db", json=json)



    ### INTERNAL

    def _post(self, path, json=None, authorize=True):
        headers = {'Authorization': 'Bearer {}'.format(self.token)} if authorize else None
        path = self.url + path
        if json is None:
            json = {}
        if json is not None and "self" in json:
            del json['self']
        json = jsonable_encoder(json)
        json["project_id"] = self.project_id
        return requests.post(path, headers=headers, json=json, cert=self.cert)



from fastapi import APIRouter

from pandahub.api.internal import settings
from pandahub.api.internal.users import fastapi_users

router = APIRouter(
    prefix="/users",
    tags=["users"]
)

router.include_router(
    fastapi_users.get_users_router(
        requires_verification=settings.EMAIL_VERIFICATION_REQUIRED),
)



import pandapower.networks as nw
from pandahub import PandaHubError
import pandapower as pp
import pytest

def test_network_io(ph):
    ph.set_active_project("pytest")

    net1 = nw.mv_oberrhein()
    name1 = "oberrhein_network"

    net2 = nw.simple_four_bus_system()
    pp.create_bus(net2, vn_kv=20, index=10) #check non-consecutive indices
    name2 = "simple_network"

    #we check storing two different networks consecutively to ensure the nets are properly separated
    for net, name in [(net1, name1), (net2, name2)]:
        if ph.network_with_name_exists(name):
            ph.delete_net_from_db(name)
        
        assert ph.network_with_name_exists(name) == False
            
        ph.write_network_to_db(net, name)
        assert ph.network_with_name_exists(name) == True
        
        net_loaded = ph.get_net_from_db(name)
        
        pp.runpp(net)
        pp.runpp(net_loaded)
        
        assert pp.nets_equal(net, net_loaded, check_only_results=True)
    
        net3 = ph.get_net_from_db(name, only_tables=["bus"])
        assert len(net3.bus) == len(net.bus)
        assert len(net3.line) == 0
        assert len(net3.load) == 0
    
    #delete first network
    ph.delete_net_from_db(name1)
    assert ph.network_with_name_exists(name1) == False

    #check that second network is still in database    
    assert ph.network_with_name_exists(name2) == True
    net2_loaded = ph.get_net_from_db(name2)
    pp.runpp(net2_loaded)
    assert pp.nets_equal(net2, net2_loaded, check_only_results=True)

def test_load_subnetwork(ph):
    ph.set_active_project("pytest")
    name = "oberrhein_network"
    
    if not ph.network_with_name_exists(name):
        net = nw.mv_oberrhein()
        ph.write_network_to_db(net, name)
    
    subnet = ph.get_subnet_from_db(name, bus_filter={"vn_kv": 110})
    expected_sizes = [("bus", 4), ("line", 0), ("trafo", 2), ("ext_grid", 2)]

    for element, size in expected_sizes:
        assert len(subnet[element]) == size
        assert len(subnet["res_" + element]) == size

    subnet = ph.get_subnet_from_db(name, bus_filter={"vn_kv": 110},
                    include_results=False)
    for element, size in expected_sizes:
        assert len(subnet[element]) == size
        assert len(subnet["res_" + element]) == 0
        
    subnet = ph.get_subnet_from_db(name, bus_filter={"vn_kv": 110}, 
                                   add_edge_branches=False)
    expected_sizes = [("bus", 2), ("line", 0), ("trafo", 0), ("ext_grid", 2)]

    for element, size in expected_sizes:
        assert len(subnet[element]) == size
        assert len(subnet["res_" + element]) == size

def test_access_and_set_single_values(ph):
    ph.set_active_project("pytest")
    name = "oberrhein_network"
    
    net = nw.mv_oberrhein()
    if not ph.network_with_name_exists(name):
        ph.write_network_to_db(net, name)
    
    element = "sgen"
    parameter = "p_mw"
    index = 4
    p_mw_new = 0.1
    
    value = ph.get_net_value_from_db(name, element, index, parameter)
    assert value == net[element][parameter].at[index] 
    
    ph.set_net_value_in_db(name, element, index, parameter, p_mw_new)
    value = ph.get_net_value_from_db(name, element, index, parameter)
    assert value == p_mw_new

    ph.delete_net_element(name, element, index)
    with pytest.raises(PandaHubError):
        ph.get_net_value_from_db(name, element, index, parameter)
    net = ph.get_net_from_db(name)    
    assert index not in net[element].index
    
  
if __name__ == '__main__':
    from pandahub import PandaHub

    ph = PandaHub(connection_url="mongodb://localhost:27017")

    project_name = "pytest"
    
    if ph.project_exists(project_name):
        ph.set_active_project(project_name)
        ph.delete_project(i_know_this_action_is_final=True)
    
    ph.create_project(project_name)
    ph.set_active_project(project_name)
    
    name = "oberrhein_network"
    
    net = nw.mv_oberrhein()
    if not ph.network_with_name_exists(name):
        ph.write_network_to_db(net, name)
    
    element = "sgen"
    parameter = "p_mw"
    index = 4
    p_mw_new = 0.1
    
    value = ph.get_net_value_from_db(name, element, index, parameter)
    assert value == net[element][parameter].at[index] 
    
    ph.set_net_value_in_db(name, element, index, parameter, p_mw_new)
    value = ph.get_net_value_from_db(name, element, index, parameter)
    assert value == p_mw_new

    ph.delete_net_element(name, element, index)
    with pytest.raises(PandaHubError):
        ph.get_net_value_from_db(name, element, index, parameter)
    net = ph.get_net_from_db(name)    
    assert index not in net[element].index
    
    

    


from typing import Optional

from fastapi import Depends, Request
from fastapi_users import BaseUserManager, FastAPIUsers
from fastapi_users.authentication import AuthenticationBackend, BearerTransport
from fastapi_users.authentication.strategy.db import AccessTokenDatabase, DatabaseStrategy
from fastapi_users.db import MongoDBUserDatabase

from ..internal import settings
from ..internal.db import get_user_db, get_access_token_db
from ..internal.models import User, UserCreate, UserDB, UserUpdate, AccessToken
from ..internal.toolbox import send_password_reset_mail, send_verification_email


class UserManager(BaseUserManager[UserCreate, UserDB]):
    user_db_model = UserDB
    if settings.SECRET is None:
        raise UserWarning("You must specify a SECRET in the environment variables or .env file")
    reset_password_token_secret = settings.SECRET
    verification_token_secret = settings.SECRET

    async def on_after_register(self, user: UserDB, request: Optional[Request] = None):
        if (settings.EMAIL_VERIFICATION_REQUIRED):
            await self.request_verify(user)
        print(f"User {user.id} has registered.")

    async def on_after_forgot_password(
        self, user: UserDB, token: str, request: Optional[Request] = None
    ):
        await send_password_reset_mail(user, token)

    #         print(f"User {user.id} has forgot their password. Reset token: {token}")

    async def on_after_request_verify(
        self, user: UserDB, token: str, request: Optional[Request] = None
    ):
        await send_verification_email(user, token)


#         print(f"Verification requested for user {user.id}. Verification token: {token}")


async def get_user_manager(user_db: MongoDBUserDatabase = Depends(get_user_db)):
    yield UserManager(user_db)


bearer_transport = BearerTransport(tokenUrl="auth/login")


def get_database_strategy(
    access_token_db: AccessTokenDatabase[AccessToken] = Depends(get_access_token_db),
) -> DatabaseStrategy[UserCreate, UserDB, AccessToken]:
    return DatabaseStrategy(access_token_db)


auth_backend = AuthenticationBackend(
    name="jwt",
    transport=bearer_transport,
    get_strategy=get_database_strategy,
)
fastapi_users = FastAPIUsers(
    get_user_manager,
    [auth_backend],
    User,
    UserCreate,
    UserUpdate,
    UserDB,
)



import asyncio

import motor.motor_asyncio
from fastapi_users.db import MongoDBUserDatabase
from fastapi_users_db_mongodb.access_token import MongoDBAccessTokenDatabase

from pandahub.api.internal import settings
from pandahub.api.internal.models import AccessToken, UserDB

client = motor.motor_asyncio.AsyncIOMotorClient(
    settings.MONGODB_URL, uuidRepresentation="standard"
)

client.get_io_loop = asyncio.get_event_loop

db = client["user_management"]
collection = db["users"]
access_tokens_collection = db["access_tokens"]


async def get_user_db():
    yield MongoDBUserDatabase(UserDB, collection)


async def get_access_token_db():
    yield MongoDBAccessTokenDatabase(AccessToken, access_tokens_collection)



from fastapi_users import models
from fastapi_users.authentication.strategy.db import BaseAccessToken


class User(models.BaseUser):
    pass


class UserCreate(models.BaseUserCreate):
    pass


class UserUpdate(models.BaseUserUpdate):
    pass


class UserDB(User, models.BaseUserDB):
    pass


class AccessToken(BaseAccessToken):
    pass



from fastapi_mail import FastMail, MessageSchema, ConnectionConfig

from pandahub.api.internal import settings

email_conf = ConnectionConfig(
    MAIL_USERNAME=settings.MAIL_USERNAME,
    MAIL_PASSWORD=settings.MAIL_PASSWORD,
    MAIL_FROM=settings.MAIL_USERNAME,
    MAIL_PORT=settings.MAIL_PORT,
    MAIL_SERVER=settings.MAIL_SMTP_SERVER,
    MAIL_TLS=settings.MAIL_TLS,
    MAIL_SSL=settings.MAIL_SSL
)

fast_mail = FastMail(email_conf)


async def send_password_reset_mail(user, token):
    to_address = user.email
    subject = "retoflow Password Reset"
    reset_link = f"{settings.PASSWORD_RESET_URL}?reset_token={token}"
    template = f"""\
Hello {user.email.split('@')[0]}, \n\n
you requested a password reset. To reset your password please click
the following link: \n\n
{reset_link}"
\n\n
If you didn't request a reset you can ignore this email. But be aware that
someone is trying to reset your password!
    """

    message = MessageSchema(
        subject=subject,
        recipients=[to_address],
        body=template,
        subtype="plain"
    )

    await fast_mail.send_message(message)


async def send_verification_email(user, token):
    to_address = user.email
    subject = "retoflow Email Verification"
    verification_link = f"{settings.EMAIL_VERIFY_URL}?verify_token={token}"
    template = f"""\
Hello {user.email.split('@')[0]}, \n\n
to complete your retoflow registration and to verify your email address please click
the following link: \n\n
{verification_link}
    """

    message = MessageSchema(
        subject=subject,
        recipients=[to_address],
        body=template,
        subtype="plain"
    )

    await fast_mail.send_message(message)



# -*- coding: utf-8 -*-
"""
Created on Tue Feb  2 11:32:07 2021

@author: julffers
"""

import pandas as pd
import pandapower as pp
import pandapipes as pps
import base64
import hashlib
import logging
import json
logger = logging.getLogger(__name__)


def get_document_hash(task):
    """
    Returns a hash value of the input task in order to generate consistent but
    unique _ids.

    Parameters
    ----------
    task : dict
        task that shall be hashed.

    Returns
    -------
    sha256 hash
        hash value that can be uses as _id.

    """
    hasher = hashlib.sha256()
    hasher.update(repr(make_task_hashable(task)).encode())
    return base64.urlsafe_b64encode(hasher.digest()).decode()

def make_task_hashable(task):
    """
    Makes a task dict hashable.
    Parameters
    ----------
    task : dict
        task that shall be made hashable.
    Returns
    -------
    TYPE
        hashable task.
    """
    if isinstance(task, (tuple, list)):
        return tuple((make_task_hashable(e) for e in task))

    if isinstance(task, dict):
        return tuple(sorted((k,make_task_hashable(v)) for k,v in task.items()))

    if isinstance(task, (set, frozenset)):
        return tuple(sorted(make_task_hashable(e) for e in task))
    return task


def add_timestamp_info_to_document(document, timeseries):
    """
    Adds some meta information to documents that containg time series.
    The document will get the additional attributes 'first_timestamp',
    'last_tiemstamp' and 'num_timestamps' (number of timestamps). This
    information can be used to calculate the timeseries' resolution, to
    check, if there are any timestamps missing or to query documents based
    on their contained tiemstamps.

    Parameters
    ----------
    document : dict
        dict, that contains the timeseries' metadata.
    timeseries : pandas.Series
        The timeseries itself with the timestamps as index.

    Returns
    -------
    document : dict
        The updated timeseries metadata document.

    """
    document["first_timestamp"] = timeseries.index[0]
    document["last_timestamp"] = timeseries.index[-1]
    document["num_timestamps"] = len(timeseries.index)
    document["max_value"] = timeseries.max().item()
    document["min_value"] = timeseries.min().item()
    return document

def convert_timeseries_to_subdocuments(timeseries):
    """
    Converts a timeseries to a list of dicts. Every dict represents one timestep
    and contains the keys 'timestamp' and 'value' as well as the according values.

    Parameters
    ----------
    timeseries : pandas.Series
        A timeseries with the timestamps as index.

    Returns
    -------
    subdocuments : list
        List of timestep dictionaries.

    """
    subdocuments = []
    for timestamp, value in list(timeseries.items()):
        subdocuments.append({"timestamp": timestamp,
                             "value": value})
    return subdocuments

def create_timeseries_document(timeseries, data_type, element_type=None,
                               netname=None, element_index=None, **kwargs):
    """
    Creates a document that contains timeseries metadata as well as the timeseries
    itself. Uses the function 'add_timestamp_info_to_document' to add information
    about the embedded timeseries and the function 'convert_timeseries_to_subdocuments'
    to convert the input timeseries to a list of subdocuments, that can be accessed
    with the attribute "timeseries_data". The element_type and data_type of the
    timeseries are required metadata, netname and element_index are optional.
    Additionally, arbitrary kwargs can be added.

    By default, the document will have an _id generated based on its metadata
    using the function get_document_hash but an _id provided by the user as a
    kwarg will not be overwritten.


    Parameters
    ----------
    timeseries : pandas.Series
        A timeseries with the timestamps as index.
    element_type : str
        Kind of element the timeseries belongs to (load/sgen).
    data_type : str
        Type and unit the timeseries values' are given in. The recommended format
        is <type>_<unit> (e.g. p_mw).
    netname : str, optional
        Name of the network the timeseries belongs to. Is only added to the
        document if any value is specified. The default is None.
    element_index : int (could also be str, but int is recommended), optional
        The index of the element the timeseries belongs to. Is only added to the
        document if any value is specified. The default is None.
    **kwargs :
        Any additional metadata that shall be added to the document.

    Returns
    -------
    document : dict
        dict, representing a timeseries.

    """
    document = {"data_type": data_type}
    if element_type is not None:
        document["element_type"] = element_type
    if netname is not None:
        document["netname"] = netname
    if element_index is not None:
        document["element_index"] = element_index
    document = add_timestamp_info_to_document(document, timeseries)
    document = {**document, **kwargs}
    if not "_id" in document: # IDs set by users will not be overwritten
        document["_id"] = get_document_hash(document)
    document["timeseries_data"] = convert_timeseries_to_subdocuments(timeseries)
    return document

def convert_dataframes_to_dicts(net, _id):
    dataframes = {}
    other_parameters = {}
    types = {}
    for key, data in net.items():
        if key.startswith("_"):
            continue
        if isinstance(data, pd.core.frame.DataFrame):
            if data.empty:
                continue
            # convert pandapower objects in dataframes to dict
            if "object" in net[key].columns:
                net[key]["object"] = net[key]["object"].apply(lambda obj: obj.to_dict())
            net[key]["index"] = net[key].index
            net[key]["net_id"] = _id
            dataframes[key] = net[key].to_dict(orient="records")
            net[key].drop(columns=["index", "net_id"], inplace=True)
            types[key] = {column: str(dtype) for column, dtype
                                       in net[key].dtypes.items()}
        else:
            try:
                json.dumps(data)
            except:
                print("Data in net[{}] is not JSON serializable and was therefore omitted on import".format(key))
            else:
                other_parameters[key] = data
    return dataframes, other_parameters, types
